'''
Created on Jul 10, 2019

@author: shuangyinli
'''
import sys
import re
import numpy as np
from scipy.special import gammaln, psi,polygamma

import random

dictionary = []

TopicNo = 0
Elogbeta = list()


def dirichlet_expectation(alpha):
    """
    For a vector theta ~ Dir(alpha), computes E[log(theta)] given alpha.
    """
    if (len(alpha.shape) == 1):
        return(psi(alpha) - psi(np.sum(alpha)))
    return(psi(alpha) - psi(np.sum(alpha, 1))[:, np.newaxis])


def removecontaindot(abstract):
    abstract = re.sub("[^A-Za-z.]", ' ', abstract)
    ###remove a single word
    abstract=re.sub(' [a-zA-Z] ', ' ', abstract)
    abstract=re.sub('^[ ]*[a-zA-Z] ', ' ', abstract) 
    abstract=re.sub(' [a-zA-Z][ ]*$', ' ', abstract)
    abstract =  ' '.join(abstract.split())
    abstract = abstract.lower()
    return abstract

def getDictionary(vocab_file):
    vocab = open(vocab_file, 'r', encoding = "utf-8").readlines()
    for line in vocab:
        dictionary.append(line.strip().rstrip().split(":")[1])
    pass

def getbeta(beta_file):
    topicDis= []
    betafile = open(beta_file, "r", encoding = "utf-8")
    
    betalist = []
    for topic in betafile:
        wordlist =[float(m) for m in topic.split()]
        betalist.append(wordlist)
        id_list = np.argsort(-np.array(wordlist))
        word_ranks = [dictionary[id] for id in id_list]
        topicDis.append(word_ranks)
        
    Elogbeta = np.array(betalist, dtype ="float")
    
    return topicDis,len(Elogbeta),Elogbeta

def codebiRATM(text):
    
    document = []
    lda_doc = {}
    
    text_remove = removecontaindot(text)
    for sen in text_remove.rstrip().lstrip().split("."):
        sent = sen.rstrip().lstrip().split()
        sentence = {}
        #wordids = list()
        #wordcts = list()
        for word in sent:
            if word in dictionary:
                wordid = dictionary.index(word)
                if wordid not in sentence:
                    sentence.setdefault(wordid)
                    sentence[wordid] = 1
                else:
                    sentence[wordid] += 1
                #for lda
                if wordid not in lda_doc:
                    lda_doc.setdefault(wordid)
                    lda_doc[wordid] =1
                else:
                    lda_doc[wordid] +=1
                    
        #wordids.append(list(sentence.keys()))
        #wordcts.append(list(sentence.values()))
        if len(list(sentence.keys())) > 0:
            document.append((list(sentence.keys()),list(sentence.values())))
    
    lda_wordids = list(lda_doc.keys())
    lda_wordcts = list(lda_doc.values())
    
    return document, (lda_wordids, lda_wordcts)

def getDescentXi(bi_neighbor, wordidlist, wordcntlist, xi, neighbortopics,phinorm):
    num_neighbors = bi_neighbor
    
    descent_xi = np.zeros(num_neighbors, dtype = "float")
    
    sigma_pi = 0.1 * num_neighbors
    sigma_xi = sum(xi)
    
    for i in range(num_neighbors):
        descent_xi[i] = polygamma(1,xi[i]) * (0.01 - xi[i])
        descent_xi[i] -= polygamma(1,sigma_xi) * (sigma_pi - sigma_xi)
        
    sum_log_phi = np.zeros(TopicNo, dtype = "float")
    sum_gamma_array = np.zeros(TopicNo, dtype = "float")
    for k in range(TopicNo):
        for i in range(num_neighbors):
            sum_log_phi[k] += neighbortopics[i][k] * xi[i]
        for n in range(len(wordidlist)):
            sum_gamma_array[k] += phinorm[n][k] * wordcntlist[n]
        
    temp =0
    for i in range(num_neighbors):
        for k in range(TopicNo):
            temp += neighbortopics[i][k] * sigma_xi
            temp -= sum_log_phi[k]
            temp = sum_gamma_array[k] * temp / (sigma_xi * sigma_xi)
            descent_xi[i] += temp
    
    return descent_xi
            
def getxifunc(bi_neighbor, wordidlist, wordcntlist, neighbortopics, xi, phinorm):
    xi_function_value = 0.0
    num_neighbors = bi_neighbor

    sigma_xi = sum(xi)
    
    sum_log_theta = np.zeros(TopicNo, dtype = "float")
    
    for i in range(num_neighbors):
        xi_function_value += (0.01 - xi[i]) * (psi(xi[i]) - psi(sigma_xi)) + gammaln(xi[i])
    
    xi_function_value -= gammaln(sigma_xi)
    
    for k in range(TopicNo):
        for i in range(num_neighbors):
            sum_log_theta[k] += neighbortopics[i][k] * xi[i] / sigma_xi
    
    for n in range(len(wordidlist)):
        for k in range(TopicNo):
            xi_function_value += sum_log_theta[k] * phinorm[n][k] * wordcntlist[n]
            pass
        pass
    
    return xi_function_value

def generate_attentions_topics(G0_or_not, neighbor, document, doc_topics, doc_phinorm, sentences_topics, sentences_phinorms):
    sen_no = len(document)
    bi_neighbor = 2*neighbor
    attentions = []
    sentences_topics_biRATM = np.zeros([sen_no, TopicNo], dtype="float")
    neighborTopicMatric = np.zeros([sen_no+bi_neighbor, TopicNo], dtype="float")
    for i in range(neighbor):
        neighborTopicMatric[i] = doc_topics.copy()
        neighborTopicMatric[neighbor+sen_no+i] = doc_topics.copy()
    
    for s in range(sen_no):
        neighborTopicMatric[neighbor + s] = sentences_topics[s].copy()

    for s in range(sen_no):
        neighbortopics = np.zeros([bi_neighbor, TopicNo], dtype="float")
        for i in range(neighbor):
            neighbortopics[i] = neighborTopicMatric[s+neighbor - i -1].copy()
        for i in range(neighbor):
            neighbortopics[i+neighbor] = neighborTopicMatric[s+neighbor+i +1].copy()
        
        wordidlist = document[s][0]
        wordcntlist = document[s][1]
        
        if G0_or_not is True:
            neighbortopics = np.insert(neighbortopics, 0, values = doc_topics, axis =0)
            s_xi = generate_xi(bi_neighbor+1, wordidlist, wordcntlist, neighbortopics, sentences_phinorms[s])
            s_xi = s_xi / sum(s_xi)
            sentences_topics_biRATM[s] = np.dot(s_xi, neighbortopics) + sentences_topics[s]
        else:
            s_xi = generate_xi(bi_neighbor, wordidlist, wordcntlist, neighbortopics, sentences_phinorms[s])
            s_xi = s_xi / sum(s_xi)
            sentences_topics_biRATM[s] = np.dot(s_xi, neighbortopics) + sentences_topics[s]
        
        # if     G0_or_not is True, the first row in s_xi is the attention of the doc topics.
        
        attentions.append(s_xi)

    return np.array(attentions), sentences_topics_biRATM

def generate_topics_by_attentions(G0_or_not, neighbor, index_focus, s_xi, doc_topics, sentences_topics):
    sen_no = len(document)
    bi_neighbor = 2*neighbor
    
    select_sentence_topics = np.zeros(TopicNo, dtype="float")
    neighborTopicMatric = np.zeros([sen_no+bi_neighbor, TopicNo], dtype="float")
    for i in range(neighbor):
        neighborTopicMatric[i] = doc_topics.copy()
        neighborTopicMatric[neighbor+sen_no+i] = doc_topics.copy()
    for s in range(sen_no):
        neighborTopicMatric[neighbor + s] = sentences_topics[s].copy()
    
    neighbortopics = np.zeros([bi_neighbor, TopicNo], dtype="float")
    for i in range(neighbor):
        neighbortopics[i] = neighborTopicMatric[s+neighbor - i -1].copy()
    for i in range(neighbor):
        neighbortopics[i+neighbor] = neighborTopicMatric[s+neighbor+i +1].copy()
        
    if G0_or_not is True:
        neighbortopics = np.insert(neighbortopics, 0, values = doc_topics, axis =0)
        s_xi = s_xi / sum(s_xi)
 
        select_sentence_topics = np.dot(s_xi, neighbortopics) + sentences_topics[index_focus]
    else:
        s_xi = s_xi / sum(s_xi)
        select_sentence_topics = np.dot(s_xi, neighbortopics) + sentences_topics[index_focus]

    return select_sentence_topics



def generate_xi(bi_neighbor, wordidlist, wordcntlist, neighbortopics, phinorm):
    num_neighbors = bi_neighbor
    xi = np.zeros(num_neighbors, dtype = "float")
    global learn_rate
    for i in range(num_neighbors):
        xi[i] = random.random()*100
    
    z = getxifunc(num_neighbors, wordidlist, wordcntlist, neighbortopics, xi, phinorm)
    
    last_xi = np.zeros(num_neighbors, dtype = "float")
    eps = 10000
    iteration_num = 0
    while (eps > 0.001 or iteration_num < 10):
        last_z = z
        last_xi = xi.copy()
        descent_xi = getDescentXi(num_neighbors, wordidlist, wordcntlist, xi, neighbortopics,phinorm)
        
        has_neg_value_flag = False
        
        for i in range(num_neighbors):
            xi[i] += learn_rate * descent_xi[i];
            if xi[i] < 0:
                has_neg_value_flag = True
        
        z = getxifunc(num_neighbors, wordidlist, wordcntlist, neighbortopics, xi, phinorm)
        
        if has_neg_value_flag ==True or last_z > z:
            learn_rate *= 0.5
            last_z = z
            eps = 10000
            xi = last_xi.copy()
        else:
            eps = np.linalg.norm(xi - last_xi)
        
        iteration_num +=1
    
    return xi

def generate_topic_sentence(lda_wordids, lda_wordcts):
    # the input is formated as ([id1,id2,...idn],[3,4,..,1])
    #not [wordid:count,wordid:count,...,wordid:count]
    
    gamma = 1*np.random.gamma(100., 1./100., TopicNo)
    Elogtheta = dirichlet_expectation(gamma)
    expElogtheta = np.exp(Elogtheta)
    
    ids = lda_wordids
    cts = lda_wordcts
    
    expElogbeta = np.exp(Elogbeta)[:,ids]
    phinorm = np.dot(expElogtheta, expElogbeta) + 1e-100
    phi = expElogbeta.T * expElogtheta # shape: N*K
    
    for it in range(0, 100):
        lastgamma = gamma
        gamma = 0.001 + sum((cts * phi.T / phinorm).T)
        Elogtheta = dirichlet_expectation(gamma)
        expElogtheta = np.exp(Elogtheta)
        phi = expElogbeta.T * expElogtheta
        phinorm = np.dot(expElogtheta, expElogbeta) + 1e-100
        meanchange = np.mean(abs(gamma - lastgamma))
        if (meanchange < 0.001):
            break
    
    norm_phi = (phi.T / phinorm).T
    
    return gamma, norm_phi

def genetate_topic_document(sentence_list):
    
    num_sentence_list = len(sentence_list)
    
    gamma = 1*np.random.gamma(100., 1./100., (len(sentence_list), TopicNo))
    Elogtheta = dirichlet_expectation(gamma)
    expElogtheta = np.exp(Elogtheta)
    sentences_phinorm = []
    
    for d in range(num_sentence_list):
        ids = sentence_list[d][0]
        cts = sentence_list[d][1]
        
        gammad = gamma[d, :]
        Elogthetad = Elogtheta[d, :]
        expElogthetad = expElogtheta[d, :]
        expElogbetad = np.exp(Elogbeta)[:,ids]
        
        phinorm = np.dot(expElogthetad, expElogbetad) + 1e-100
        phi = expElogbetad.T * expElogthetad # shape: N*K

        for it in range(0, 100):
            lastgamma = gammad
            gammad = 0.001 + sum((cts * phi.T / phinorm).T)
            Elogthetad = dirichlet_expectation(gammad)
            expElogthetad = np.exp(Elogthetad)
            phi = expElogbetad.T * expElogthetad
            phinorm = np.dot(expElogthetad, expElogbetad) + 1e-100
            meanchange = np.mean(abs(gammad - lastgamma))
            if (meanchange < 0.001):
                    break
        
        gamma[d, :] = gammad
        
        norm_phi = (phi.T / phinorm).T # shape:   N*K
        sentences_phinorm.append(norm_phi)
    
    return gamma, sentences_phinorm

if __name__ == '__main__':
    if (len(sys.argv) != 3):
        print("usage: python manualAttentions.py <dictionary file><semanic probability over dictionary file(final.beta)> \n")
        print(" \n")
        sys.exit(1)
         
    dictionary_file = sys.argv[1]
    beta_file = sys.argv[2]
     
    # read 
    print("Begin to read dictionary.")
    getDictionary(dictionary_file)
    topicDis,TopicNo, Elogbeta = getbeta(beta_file)
     
    while True:
        doctext = str(input("Enter one doc with several sentences, the sentences are splited by dots:"))
        
        doctext = removecontaindot(doctext)
        
        if doctext is "":
            print("exit")
            exit()
        if len(doctext.split(".")) < 3:
            print("The number of sentences in the doc is less than 3, please provide more sentencs")
            continue

        learn_rate = 1
        
        print("begin to precess the input doc, please wait...")
        
        document, (lda_wordids, lda_wordcts) = codebiRATM(doctext)
        
        print("The length of the input document is " +str(len(document)))
    
        neighbor = int(input("Please set the neighbor number for bi-RATM (less than length of the document) :"))
    
        if neighbor > len(document) or neighbor < 0:
            print("You've required too many neighbors! not happy...")
            exit()
        
        
        G0_or_not = input("Do you want consider the host document as one of the neighbors? True or False: ")
        
        if G0_or_not == "True" or G0_or_not == "true" or G0_or_not == "t" or G0_or_not == "T":
            G0_or_not = True
        else:
            G0_or_not = False
        
        
        print("begin to compute the topics, please wait...")
        
        sentences_topics, sentences_phinorms = genetate_topic_document(document)
        
        doc_topics, doc_phinorm = generate_topic_sentence(lda_wordids, lda_wordcts)
        
        print("begin to compute the attentions, please wait...")
        # if G0_or_not is True, the first row in s_xi is the attention of the doc topics.
        attentions, sentences_topics_biRATM = generate_attentions_topics(G0_or_not, neighbor, document, doc_topics, doc_phinorm, sentences_topics, sentences_phinorms)
        
        print("Done! \n")
                
        
        while True:
            focus_sentence =input("Enter which sentence you want to generate:")
            
            if focus_sentence == "":
                exit()
            
            if int(focus_sentence) <= 0 or int(focus_sentence) > len(document):
                print("Please enter a number from 1 to "+ str(len(document)) + "\n")
                continue
            elif int(focus_sentence) > 0 and int(focus_sentence) < len(document):
                
                if G0_or_not is True:
                    focus = int(focus_sentence)
                    
                    print("\nAttention of the host document:" + str(attentions[focus-1][0]))
                    print("Attentions of the preceding sentences:" + str(attentions[focus-1][1:neighbor+1]))
                    print("Attentions of the following sentences:" + str(attentions[focus-1][neighbor+1:]) + "\n")
                    
                    print("The top 5 topics (ranked) of the selected sentences described by words: \n")
                    topics = sentences_topics_biRATM[focus -1]
                    topicsid = np.argsort(-np.array(topics))[:5]
                    for id in topicsid:
                        resl = topicDis[id]
                        print(str(id) + ": " + str(resl[:10]))
                    print("\n")
                else:
                    # people would like to count from 1
                    focus = int(focus_sentence)
                    
                    print("\nAttentions of the preceding sentences:" + str(attentions[focus-1][0:neighbor]))
                    print("Attentions of the following sentences:" + str(attentions[focus-1][neighbor:]) + "\n")
                    
                    print("The top 5 topics (ranked) of the selected sentences described by words: \n")
                    topics = sentences_topics_biRATM[focus -1]
                    topicsid = np.argsort(-np.array(topics))[:5]
                    for id in topicsid:
                        resl = topicDis[id]
                        print(str(id) + ": " + str(resl[:10]))
                    print("\n")
                
                while True:
                    
                    print("Now please input the attentions manually:\n")
                    
                    if G0_or_not is True:
                        attention_doc = float(input("Enter the attention of the host docment between ([0,1]): "))
                        print("Now please input the attentions of the neighbored sentences (example: 0.1 0.3 0.4 0.2:), float values between [0,1], splited by space")
                        attention_list = input(str(2*neighbor) +" float values : \n")
                        attentionlist = []
                        attentionlist.append(attention_doc)
                        for at in attention_list.split():
                            attentionlist.append(float(at))
                        
                        attentions_array = np.array(attentionlist)
                        
                        print("The top 5 topics (ranked) of the selected sentences with the manual attentions: \n")
                        focus_index = int(focus_sentence) -1
                        
                        topics = generate_topics_by_attentions(G0_or_not, neighbor, focus_index, attentions_array, doc_topics, sentences_topics)
                        topicsid = np.argsort(-np.array(topics))[:5]
                        for id in topicsid:
                            resl = topicDis[id]
                            print(str(id) + ": " + str(resl[:10]))
                        print("\n")
                        
                        pass
                    else:
                        print("Now please input the attentions of the neighbored sentences (example: 0.1 0.3 0.4 0.2:), float values between [0,1], splited by space")
                        attention_list = input(str(2*neighbor) +" float values : \n")
                        attentionlist = []
                        for at in attention_list.split():
                            attentionlist.append(float(at))
                        
                        attentions_array = np.array(attentionlist)
                        
                        print("The top 5 topics (ranked) of the selected sentences with the manual attentions: \n")
                        focus_index = int(focus_sentence) -1
                        
                        topics = generate_topics_by_attentions(G0_or_not, neighbor, focus_index, attentions_array, doc_topics, sentences_topics)
                        topicsid = np.argsort(-np.array(topics))[:5]
                        for id in topicsid:
                            resl = topicDis[id]
                            print(str(id) + ": " + str(resl[:10]))
                        print("\n")
                        
                        pass
                    pass
                
                    tag = input("Do you want to continue: y or n:")
                    if tag == "y" or tag == "Y" or tag=="yes" or tag == "Yes" or tag=="YES":
                        continue
                    else:
                        break
            
            else:
                break
        
        